#include "AmbOccl.h"

#include "Mesh.h"
#include "Random.h"
#include "Ray.h"
#include "util.h"
#include "Vec3.h"
#include "InputHandler.h"
#include "Hemicube.h"

#include <vector>
#include <string>
#include <map>
#include <fstream>
#include <iostream>
#include <cmath>

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

using namespace ambenv;

// -- Some local function definitions.
int InitGL( Scene & scene, unsigned size );
void DestroyGL( Scene & scene );
void DrawMesh( const Mesh * mesh );

extern unsigned MeshTriIntersect;
extern unsigned MeshTriCacheHits;
extern unsigned MeshTriCacheMisses;
extern unsigned MeshTriCacheSize;

const std::string PVO_PATH( "data/pvo/" );
const std::string PVO_EXT( ".pvo" );
const std::string PVO_HC_PATH( "data/pvo/hc/" );

bool ambenv::write_pv( Mesh * mesh, LogMessage & log, const std::string & path  )
{   
  std::string filename = path + mesh->getName() + PVO_EXT;
  

  // -- Do not write the file if it already exists.
  std::ifstream filecheck( filename.c_str(), std::ios_base::in );
  if( filecheck.is_open() ) {
    log.lm_fprintf( lm_warn, "Cannot overwrite file (%s).\n", filename.c_str() );
    return( false );
  }
  
  
  std::ofstream outfile( filename.c_str(), std::ios_base::out );
  if( !outfile.is_open() ) {
    log.lm_fprintf( lm_err, "Cannot write file (%s).\n", filename.c_str() );
    return( false );
  }
  
  outfile << "# -- " << filename << "\n";
  outfile << "# -- " << mesh->numVertices() << " vertices, ";
  outfile << mesh->numTriangles() << " triangles\n"; 
  
  // -- For each vertex.
  for( unsigned i=0; i<mesh->numVertices(); i++ ) {
    MeshVertex vertex = mesh->getVertex( i );
    outfile << vertex.occl() << " ";
    outfile << vertex.bentNormal().x() << " ";
    outfile << vertex.bentNormal().y() << " ";
    outfile << vertex.bentNormal().z() << "\n";
  }
  
  outfile << "# -- EOF\n";
  outfile.close();
  
  return( true );
}

bool ambenv::read_pv( Mesh * mesh, LogMessage & log, const std::string & path )
{
  
  std::string filename = path + mesh->getName() + PVO_EXT;
  
  std::ifstream infile( filename.c_str() );
  if( !infile.is_open() ) {
    log.lm_fprintf( lm_warn, "Cannot read file (%s).\n", filename.c_str() );
    return( false );
  }
  
  std::vector<std::pair<Real, Vec3> > buffer;
  buffer.reserve( mesh->numVertices() );
  
  InputHandler handler( infile );
  while( handler.getLine() == 1 ) {
    std::vector<std::string> svec = handler.getArgs();
    if( svec.size() != 4 ) {
      log.lm_fprintf( lm_err, 
        "Error reading file (%s:%u). Expected %u tokens, got %u.\n",
        filename.c_str(), handler.getLineNumber(), 
        4, svec.size() );

      std::cerr << "(" << svec[4] << ")" << "\n";

      return( false );
    }
    
    Real inputvals[4] = { 0.0, 0.0, 0.0, 0.0 };
    for( unsigned i=0; i<svec.size(); i++ ) {
      double val = strtod( svec[i].c_str(), (char **)NULL );
      if( ERANGE == errno ) {
        log.lm_fprintf( lm_err,
          "Error reading file (%s%u). Cannot convert (%s) to Real.\n",
          filename.c_str(), handler.getLineNumber(),
          svec[i].c_str() );
        
        return( false );
      }
      inputvals[i] = (Real)val;
    }
    
    // -- Finally, we've got it all.
    buffer.push_back( std::pair<Real, Vec3>
      ( inputvals[0], 
      Vec3( inputvals[1], inputvals[2], inputvals[3] ) ) );
    
  }
  
  if( buffer.size() != mesh->numVertices() ) {
    log.lm_fprintf( lm_err,
      "Error reading file (%s:%u). Read %u vertices, expected %u.\n",
      filename.c_str(), handler.getLineNumber(),
      buffer.size(), mesh->numVertices() );
    
  } else {
    
    // -- For each vertex.
    for( unsigned i=0; i<mesh->numVertices(); i++ ) {
      MeshVertex vertex = mesh->getVertex( i );
      vertex.occl( buffer[i].first );
      vertex.bentNormal( Vec3( buffer[i].second.x(),
        buffer[i].second.y(),
        buffer[i].second.z() ) );
      mesh->setVertex( i, vertex );
    }
  }
  
  infile.close();

  return( true );

}

void ambenv::generate_pv( Scene & scene )
{
  const unsigned NumSamples = scene.samples;
  MeshTriCacheSize = 50;

  scene.log.lm_fprintf( lm_info, "# -- Generating per-vertex occlusion values.\n" );
  scene.log.lm_fprintf( lm_info, "# -- Generating sample distribution: %u\n", NumSamples );
  
#ifndef WIN32
  time_t begin = time( NULL );
#endif

  std::vector<std::vector<Vec3> > distrib( NumSamples );
  {
    Random rand;
    for( unsigned i=0; i<NumSamples; i++ ) {
      distrib[i].resize( NumSamples );
      for( unsigned j=0; j<NumSamples; j++ ) {
        Real r = (Real)sqrt( rand.getReal() );
        Real theta = 2.0 * REAL_PI * rand.getReal();
        Real x = r * (Real)cos( theta );
        Real z = r * (Real)sin( theta );
        Real y = (Real)sqrt( 1.0 - x*x - z*z );
        distrib[i][j] = Vec3( x, y, z );
      }
    }
  }  

  // -- For each Mesh object.
  std::map<std::string, Object *>::iterator itr = scene.objects.begin();
  for( ; itr != scene.objects.end(); ++itr ) {

    Mesh * mesh = (Mesh *)(itr->second);


    // -- First try to read in a file with baked values.
    if( read_pv( mesh, scene.log, PVO_PATH ) ) {
      continue;
    }

    scene.log.lm_fprintf( lm_info, "# -- %u vertices, %u triangles\n", 
			  mesh->numVertices(), mesh->numTriangles() );
    scene.log.lm_fprintf( lm_info, "# -- Cache Size, Hits, Misses, Intersect Calls\n" );
    
    // -- For each vertex.
    for( unsigned i=0; i<mesh->numVertices(); i++ ) {
      
      MeshVertex vertex = mesh->getVertex( i );
      
      // -- Find rotation matrix R that rotates
      // -- the hemisphere of samples into a coordinate
      // -- system defined by the normal.
      Matrix4x4 R;
      {
        Vec3 up( 0.0, 1.0, 0.0 );
        Vec3 n = vertex.n();
        
        if( !up.approximates( n ) ) {
          
          if( up.approximates( n*-1.0 ) ) {	  
            R( 1, 1, -1.0 );
            R( 1, 2, 0.0 );
            R( 2, 2, -1.0 );
          } else {
            Vec3 rgt = n.cross( up );
            Vec3 fwd = n.cross( rgt );
            
            {
              Real mag = rgt.length();
              if( mag != 0 ) {
                rgt /= mag;
              }
              mag = fwd.length();
              if( mag != 0 ) {
                fwd /= mag;
              }
            }
            
            R( 0, 0, rgt.x() );
            R( 1, 0, rgt.y() );
            R( 2, 0, rgt.z() );
            R( 0, 1, n.x() );
            R( 1, 1, n.y() );
            R( 2, 1, n.z() );
            R( 0, 2, fwd.x() );
            R( 1, 2, fwd.y() );
            R( 2, 2, fwd.z() );
            
          }
          
          // -- Sanity check.
          assert( n.approximates( R.multVec( up, false ) ) );
          
        }
      }
      
      Real accum = 0.0;

      std::vector<Vec3> contribRays;
      Vec3 meanRay;

      for( unsigned j=0; j<NumSamples; j++ ) {
	for( unsigned k=0; k<NumSamples; k++ ) {

	  Vec3 dir = R.multVec( distrib[j][k], false );

	  Ray ray( vertex.v(), dir );
	  
	  if( !mesh->intersect( ray, 0.1 ) ) {
	    accum += 1.0;
	    contribRays.push_back( dir );
	    meanRay += dir;
	  }
	}
      }

      // -- Now, get the unit length mean ray that did not intersect.
      {
        Real mag = meanRay.length();
        if( mag != 0.0 ) {
          meanRay.normalize();
        } else {
          meanRay = Vec3( 0.0, 1.0, 0.0 );
        }
      }
      
      // -- Find the median ray. The one that is closest to the mean ray.
      Vec3 medianRay = vertex.n();
      Real minv = REAL_MAX;
      std::vector<Vec3>::iterator cr_itr = contribRays.begin();
      for( ; cr_itr != contribRays.end(); ++cr_itr ) {
        Real magSq = (meanRay - *cr_itr).lengthSq();
        if( magSq < minv ) {
          medianRay = *cr_itr;
          minv = magSq;
        }
      }

      // -- That's our bent normal.
      vertex.bentNormal( medianRay );

      // -- Set the occlusion value too.
      accum /= (NumSamples * NumSamples);
      vertex.occl( accum );

      
      mesh->setVertex( i, vertex );
      
      if( i && !(i % 100)) {
	scene.log.lm_printf( "%.2f complete, %u\n",
			     (float)i/(float)mesh->numVertices(),
			     MeshTriIntersect );
      }
    }

    // -- Write out the baked values.
    write_pv( mesh, scene.log, PVO_PATH );

  }
#ifndef WIN32
  time_t end = time( NULL );
#endif

  scene.log.lm_printf( "%.2f %u\n", 1.0, MeshTriIntersect );
#ifndef WIN32
  scene.log.lm_printf( "Completed in %u seconds.\n", end-begin );
#endif

  if( MeshTriIntersect > 0 ) {
    scene.log.lm_fprintf( lm_info, "%u %u %u %u\n", 
			  MeshTriCacheSize, MeshTriCacheHits, 
			  MeshTriCacheMisses, MeshTriIntersect ); 
  } else {
    scene.log.lm_fprintf( lm_info, 
			  "# -- Read per-vertex occlusion data from cache files.\n" ); 
  }

  scene.log.lm_fprintf( lm_info, "# -- END per-vertex occlusion generation.\n\n" );

}

void ambenv::generate_pv_hemicube( Scene & scene )
{
  int size = scene.samples;
  if( size <= 0 ) {
    size = 64;
  } else if( (size % 4) != 0 ) {
    size += size % 4;
  }

  if( InitGL( scene, size ) != 0 ) {
    return;
  }


  // -- For the raster data read from the frame buffer.
  std::vector<GLubyte> data( size * size );

  // -- Store weights associated with each pixel in frame buffer.
  std::vector<Real> weights;
  std::vector<Vec3> normals;
  
  Hemicube hcube( size );
  hcube.computeWeightsAndNormals( weights, normals );

  assert( weights.size() == (unsigned)(size * size) );
  assert( normals.size() == (unsigned)(size * size) );

  /*
  // -- Write the depth map out to file.  
  unsigned npixels = size * size;

  std::vector<unsigned char> rgb_depth( npixels * 3 );
  for( unsigned i=0; i<npixels; i++ ) {
      
    unsigned char chval = (char)(weights[i] * 256.0);

    rgb_depth[3*i] = chval;
    rgb_depth[3*i + 1] = chval;
    rgb_depth[3*i + 2] = chval;
  }

  write_rgb_tiff( "weight_map.tif", size, size, &rgb_depth[0] );
  */
  
  // -- Sum up all of the weights to normalize the occlusion value.
  Real weightSum = 0.0;
  {
    for( int i=0; i<size*size; i++ ) {
      weightSum += weights[i];
    }
  }

  // -- No depth test, no backface culling, fill front and back.
  glDisable( GL_DEPTH_TEST );
  glDisable( GL_CULL_FACE );
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  glColor3f( 1.0, 1.0, 1.0 );
  glClearColor( 0.0, 0.0, 0.0, 0.0 );
  glDepthMask( 0 );


  // -- For each Mesh object
  std::map<std::string, Object *>::iterator itr = scene.objects.begin();
  for( ; itr != scene.objects.end(); ++itr ) {

    Mesh * mesh = (Mesh *)(itr->second);

    // -- First try to read in a file with baked values.
    if( read_pv( mesh, scene.log, PVO_HC_PATH ) ) {
      continue;
    }

    // -- For each vertex
    for( unsigned i=0; i<mesh->numVertices(); i++ ) {
      MeshVertex vertex = mesh->getVertex( i ); 

      glClear( GL_COLOR_BUFFER_BIT );
      hcube.renderScene( vertex, mesh, DrawMesh );
      glFinish();

      glReadPixels( 0, 0, size, size, GL_LUMINANCE, GL_UNSIGNED_BYTE, &data[0] );

      Vec3 averageNormal = Vec3( 0, 0, 0 );
      Real accum = 0.0;
      for( int y=0; y<size; y++ ) {
        for( int x=0; x<size; x++ ) {
          if( data[size * y + x] == 0 ) {
            averageNormal += normals[size * y + x];
            accum += weights[size * y + x];
          }
        }
      }
	
      // -- Find matrix R that rotates
      // -- the hemisphere of samples into a coordinate
      // -- system defined by the normal and point.
      Matrix4x4 R;
      {
        Vec3 up( 0.0, 1.0, 0.0 );
        Vec3 n = vertex.n();
        
        if( !up.approximates( n ) ) {
          
          if( up.approximates( n*-1.0 ) ) {	  
            R( 1, 1, -1.0 );
            R( 1, 2, 0.0 );
            R( 2, 2, -1.0 );
          } else {
            Vec3 rgt = n.cross( up );
            Vec3 fwd = n.cross( rgt );
            
            {
              Real mag = rgt.length();
              if( mag != 0 ) {
                rgt /= mag;
              }
              mag = fwd.length();
              if( mag != 0 ) {
                fwd /= mag;
              }
            }
            
            R( 0, 0, rgt.x() );
            R( 1, 0, rgt.y() );
            R( 2, 0, rgt.z() );
            R( 0, 1, n.x() );
            R( 1, 1, n.y() );
            R( 2, 1, n.z() );
            R( 0, 2, fwd.x() );
            R( 1, 2, fwd.y() );
            R( 2, 2, fwd.z() );
            
          }
          
          // -- Sanity check.
          assert( n.approximates( R.multVec( up, false ) ) );
          
        }
      }


      Real mag = averageNormal.lengthSq();
      if( mag != 0.0 ) {
	mag = (Real)sqrt( mag );
	averageNormal /= mag;
      } else {
	averageNormal = vertex.n();
      }

      averageNormal = R.multVec( averageNormal );
      vertex.bentNormal( averageNormal );

      vertex.occl( accum / weightSum );
      
      mesh->setVertex( i, vertex );
      
      SDL_GL_SwapBuffers();
      
      SDL_Event event;
      while( SDL_PollEvent( &event ) ) {
        if( (event.type == SDL_QUIT) || 
	    ((event.type == SDL_KEYDOWN) &&  (event.key.keysym.sym == SDLK_ESCAPE)) ) {
          DestroyGL( scene );
          return;
        }
      }
    }

    // -- Write out the baked values.
    write_pv( mesh, scene.log, PVO_HC_PATH );

  }

  DestroyGL( scene );
}

int InitGL( Scene & scene, unsigned size )
{
  int bpp = 0;
  int rgb_size[3];
  
  if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
    scene.log.lm_fprintf( lm_err, "Could not initialize SDL: %s\n", SDL_GetError() );
    return( 1 );
  }
  
  const SDL_VideoInfo* info = SDL_GetVideoInfo();
  if( info == NULL ) {
    scene.log.lm_fprintf( lm_err, "SDL video query failed: %s\n", SDL_GetError() );
    return( 1 );
  }
  
  if( bpp == 0 ) {
    if( info->vfmt->BitsPerPixel <= 8 ) {
      bpp = 8;
    } else {
      bpp = 16;
    }
  }
  
  /* Initialize the display */
  switch( bpp ) {
  case 8:
    rgb_size[0] = 3;
    rgb_size[1] = 3;
    rgb_size[2] = 2;
    break;
  case 15:
  case 16:
    rgb_size[0] = 5;
    rgb_size[1] = 5;
    rgb_size[2] = 5;
    break;
  default:
    rgb_size[0] = 8;
    rgb_size[1] = 8;
    rgb_size[2] = 8;
    break;
  }
  
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, rgb_size[0] );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, rgb_size[1] );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, rgb_size[2] );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, bpp );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  if( SDL_SetVideoMode( size, size, bpp, SDL_OPENGL ) == NULL ) {
    scene.log.lm_fprintf( lm_err, "Could not set GL mode: %s\n", SDL_GetError() );
    SDL_Quit();
    return( 1 );
  }
  SDL_WM_SetCaption( "AmbOccl", "AmbOccl" );

  return( 0 );
}

void DestroyGL( Scene & scene ) 
{
  SDL_Quit();
}

void DrawMesh( const Mesh * mesh )
{
  assert( mesh != NULL );
      
  glBegin( GL_TRIANGLES );
  for( unsigned i=0; i<mesh->numTriangles(); i++ ) {
    MeshTriangle tri = mesh->getTriangle( i );
    for( unsigned j=0; j<3; j++ ) {
      Vec3 v = mesh->getVertex( tri.vert( j ) ).v();
      glVertex3f( v.x(), v.y(), v.z() );
    }
  }
  glEnd();  
}
