#ifndef __CAMERA_H_
#define __CAMERA_H_

#include "Nameable.h"
#include "Matrix4x4.h"
#include "Transformable.h"
#include "Vec3.h"

#include <string>

namespace ambenv {

  /*! 
   * Camera class implements view and perspective projection 
   * transformations through an interface modeled after OpenGL.
   */
  class Camera : public Transformable, public Nameable {
  public:
    Camera();
    Camera( const std::string & name );

    /*!
     * Camera functions taken from the Mesa (OpenGL) graphics library implementation,
     * http://www.mesa3d.org/, using the opengl red book for reference.
     */
    void perspective( Real myfov, Real aspect, Real mynear, Real myfar );
    
    void ortho( Real left, Real right,
		Real bottom, Real top,
		Real vnear, Real vfar );
    
    void frustum( Real left, Real right,
		  Real bottom, Real top,
		  Real vnear, Real vfar );

    void lookAt( const Vec3 & eye, const Vec3 & center, const Vec3 & up );


    // -- accessor functions
    Matrix4x4 getPersp() const;
    void setPersp( const Matrix4x4 & mat );

    Matrix4x4 getInversePersp() const;
    void setInversePersp( const Matrix4x4 & mat );

    Matrix4x4 getView() const;
    void setView( const Matrix4x4 & mat );

    Matrix4x4 getInverseView() const;
    void setInverseView( const Matrix4x4 & mat );

    Vec3 eye() const { return( m_eye ); }
    Vec3 center() const { return( m_center ); }
    Vec3 up() const { return( m_up ); }

    Real getNear() const { return( m_near ); }
    Real getFar() const { return( m_far ); }

    Real getLeft() const { return( m_left ); }
    Real getRight() const { return( m_right ); }
    Real getBottom() const { return( m_bottom ); }
    Real getTop() const { return( m_top ); }

    Real getAspect() const { return( m_aspect ); }

  private:
    Matrix4x4 m_persp;
    Matrix4x4 m_inv_persp;
    Matrix4x4 m_view;
    Matrix4x4 m_inv_view;
    
    Real m_left;
    Real m_right;
    Real m_bottom;
    Real m_top;
    Real m_near;
    Real m_far;

    Vec3 m_eye;
    Vec3 m_center;
    Vec3 m_up;

    Real m_aspect;
  };

}; /* namespace ambenv */

#endif /* __CAMERA_H_ */
